﻿//////////////////////////////////////////////
// Component.h
//
//////////////////////////////////////////////

/// Defines / Macros -------------------------

#pragma once

/// Forward declarations ---------------------

namespace nkExport
{
	class Node ;
}

namespace nkImages
{
	class Image ;
}

namespace nkWinUi
{
	enum class LAYOUT_TYPE ;

	class ComponentGenerator ;
	class InputDescriptor ;
	class Layout ;
	class ScrollBar ;
	class Theme ;

	struct ChangeEventData ;
	struct KeyEventData ;
	struct WheelEventData ;
}

/// Includes ---------------------------------

// nkWinUi
#include "../Dll/DllDefines.h"

#include "../Enums/BorderKind.h"

#include "../Inputs/HandledEventData.h"

#include "../Inputs/InputCodes.h"

#include "ComponentType.h"

// nkExport
#include <NilkinsExport/Exportable/Exportable.h>

// nkMaths
#include <NilkinsMaths/Geometry/Point.h>
#include <NilkinsMaths/Geometry/Rectangle.h>

// nkMemory
#include <NilkinsMemory/Containers/StringView.h>

// Standards
#include <string>
#include <vector>

/// Class ------------------------------------

namespace nkWinUi
{	
	class DLL_WINUI_EXPORT Component : public nkExport::Exportable
	{
		protected :
		
			// Attributes
			// Layout & theme
			Layout* _layout ;
			Theme* _theme ;

			// Tailles minimums et maximums
			std::pair<float, float> _minSize ;
			std::pair<float, float> _maxSize ;
			// Taille associée
			nkMaths::Rectangle _area ;

			// Ratio à respecter
			float _ratio ;
			// Le type de component
			COMPONENT_TYPE _componentType ;

			// Component parent
			Component* _parentComponent ;
			// Liste des fils pour pouvoir les update si besoin est
			std::vector<Component*> _children ;

			// Son nom aussi
			std::string _componentName ;
			// Label
			std::string _label ;

			// Des données utilisateurs (pour les callbacks par exemple)
			void* _data ;

			// Ordre de depth dans l'interface
			unsigned int _zDepth ;

			// Si loadé
			bool _initialized ;
			// Enabled ou pas
			bool _enabled ;
			// If should be visible on load or not
			bool _visibleOnLoad ;

		protected :

			// Utilitaires pour l'export
			virtual void _retrieveChildren (std::vector<Component*>& container) ;
		
		public :
		
			// Constructeur, destructeur
			Component () ;
			virtual ~Component () ;

			// Getters
			Layout* getLayout () const ;
			Theme* getTheme () const ;
			nkMaths::Point getPositionInParent () const ;
			nkMaths::Point getPositionInScreen () const ;
			float getWidth () const ;
			float getHeight () const ;
			nkMaths::Rectangle getAreaInParent () const ;
			nkMaths::Rectangle getAreaInScreen () const ;
			virtual nkMaths::Point getBorderExtents () const ;
			virtual float getBorderExtent (BORDER_KIND border) const ;
			virtual float getLayoutAllowedWidth () const ;
			virtual float getLayoutAllowedHeight () const ;
			virtual float getLayoutAllowedOffsetX () const ;
			virtual float getLayoutAllowedOffsetY () const ;
			bool isInitialized () const ;
			Component* getParentComponent () const ;
			COMPONENT_TYPE getComponentType () const ;
			nkMemory::StringView getComponentName () const ;
			bool getVisibility () const ;
			virtual nkMemory::StringView getLabel () const ;
			void* getData () const ;
			bool isEnabled () const ;
			bool isForegroundWindow () const ;
			virtual unsigned int getZDepth () const ;
			bool getVisibleOnLoad () const ;

			// Setters
			virtual Layout* setLayout (LAYOUT_TYPE layoutType) ;
			virtual void setTheme (Theme* theme) ;
			virtual void setPositionInParent (nkMaths::Point position) ;
			virtual void setPositionInScreen (nkMaths::Point position) ;
			virtual void setSize (float width, float height) ;
			virtual void setWidth (float width) ;
			virtual void setHeight (float height) ;
			void setMinSize (float width, float height) ;
			void setMaxSize (float width, float height) ;
			void setRatio (float widthOverHeight) ;
			virtual void setAreaInParent (const nkMaths::Rectangle& area) ;
			virtual void setAreaInScreen (const nkMaths::Rectangle& area) ;
			virtual void setParentComponent (Component* parent, bool makeItVisible = true) ;
			virtual void setLabel (const nkMemory::StringView& label) ;
			void setComponentName (const nkMemory::StringView& name) ;
			virtual void setVisibility (bool value) ;
			void setData (void* data) ;
			void setEnabled (bool value) ;
			void setVisibleOnLoad (bool value) ;
				
			// Hierarchie
			void addChild (Component* child, bool makeItVisible = true) ;
			void removeChild (Component* child) ;
			void removeChild (unsigned int index) ;
			Component* getChild (unsigned int index) ;
			unsigned int getChildIndex (Component* child) ;
			void unloadWithChildren () ;
			void loadWithChildren () ;
			void updateZDepth () ;

			// Pratiques
			void bringToForeground () ;
			void focusWindow () ;

			// Conversions pratiques
			nkMaths::Point getCoordRelativeFromAbsolute (const nkMaths::Point& absCoords) ;
			nkMaths::Point getCoordAbsoluteFromRelative (const nkMaths::Point& relCoords) ;

			// Init
			virtual void load () = 0 ;
			virtual void unload () = 0 ;

			// Update
			virtual void updateWindow () ;

			// Images
			nkImages::Image paintToImage (bool withFrame = false) ;

			// Events
			void simulateEvent (const InputDescriptor& inputDescription) ;
			// Events tenus par défaut
			void onSized () ;

			// Export / Import de base, à redéfinir par les fils s'il faut
			virtual void exportClassToTree (nkExport::Node* rootNode) override ;
			// Celle-ci est appelée lors d'un export, une classe doit la réimplémenter pour l'export, plutôt que le exportClassToTree
			virtual void exportIntrospection (nkExport::Node* rootNode) ;
			virtual void importClassFromTree (nkExport::Node* rootNode) override ;
	} ;
}